'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var functions = require('./functions-effe3ff3.js');
require('folktale/concurrency/task');
var R = require('ramda');
require('folktale/result');
require('ramda-maybe');
require('util');
var throwingFunctions = require('./throwingFunctions.js');
require('folktale/maybe');
var monadHelpers = require('./monadHelpers-3af34d56.js');
var NamedTupleMap = require('namedtuplemap');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var R__default = /*#__PURE__*/_interopDefaultLegacy(R);
var NamedTupleMap__default = /*#__PURE__*/_interopDefaultLegacy(NamedTupleMap);

/**
 * Created by Andy Likuski on 2018.12.28
 * Copyright (c) 2018 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/**
 *
 * Code modified from memoize-immutable. Memoizes a function using a flatArg version of the args
 * @param {Function} fn Function expecting any number of arguments
 * @returns {function([*], Object)} A function expecting
 * an array of the normal args for the second argument
 * an object of flattened args for the first argument
 */

var memoize = function memoize(fn) {
  var cache = new NamedTupleMap__default['default']();
  var memoized = R__default['default'].curry(function (normalArgs, flatArg) {
    if (!cache.has(flatArg)) {
      var result = R__default['default'].apply(fn, normalArgs);
      cache.set(flatArg, result);
      return result;
    }

    return cache.get(flatArg);
  }); // Give a meaningful displayName to the memoized function

  if (fn.name) {
    memoized.displayName = fn.name + 'Memoized';
  }

  return memoized;
};
/** *
 * Memomizes a function to a single argument function so that we can always NamedTupleMap for the cache.
 * In order for this to work all objects have to be flattened into one big object. This Cache won't
 * accept inner objects that have changed. So the function coverts three args like
 * {a: {wombat: 1, emu: 2}}, {b: {caracal: 1, serval: 2}}, 'hamster' to
 * {arg1.a: {wombat: 1, emu: 2}, arg2.b: {caracal: 1, serval: 2}, arg3: 'hamster}.
 * You can provide any depth of objects as arguments, but it will have performance penalties
 * Consider memoizedWith to filter out unimportant data from arguments
 * @param {Function} func A function with any number and type of args
 * @returns {Function} A function that expects the same args as func
 */


var memoized = function memoized(func) {
  return memoizedWith(function () {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return args;
  }, func);
};
/** *
 * Memomizes a function to a single argument function so that we can always NamedTupleMap for the cache.
 * In order for this to work all objects have to be flattened into one big object. This Cache won't
 * accept inner objects that have changed. So the function coverts three args like
 * {a: {wombat: 1, emu: 2}}, {b: {caracal: 1, serval: 2}}, 'hamster' to
 * {arg1.a: {wombat: 1, emu: 2}, arg2.b: {caracal: 1, serval: 2}, arg3: 'hamster}.
 * You can provide any depth of objects as arguments, but it will have performance penalties.
 * To simplify the arguments and remove incomparable things like functions, use argumentFilter.
 * @param {Function} argumentFilter Expects the same number of arguments as func and returns an equal
 * number in an array. Each argument can be filtered to remove expensive objects or functions. Make
 * sure not to filter out anything that is used by func, since it will not figure into the argument uniqueness
 * @param {Function} func A function with any number and type of args. argumentFilter must match it
 * @returns {Function} A function that expects the same args as func. This function is curried, expecting
 * whatever number of arguments func is declared with, so you can call it partially
 */

var memoizedWith = function memoizedWith(argumentFilter, func) {
  // memoize(func) returns the memoized function expecting args and the flattened obj
  var memo = memoize(func); // Returns a function expecting the args for fun

  return R__default['default'].curryN(func.length, function () {
    for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
      args[_key2] = arguments[_key2];
    }

    // Function that flattens the original args and called memoizedFunc with the single flattened arg
    return R__default['default'].compose(memo(args), functions.flattenObj, // Filter out unneeded parts of the arguments, or does nothing if argumentFilter is ...args => args
    R__default['default'].apply(argumentFilter))(args);
  });
};

var isObject = R__default['default'].compose(R__default['default'].equals('Object'), R__default['default'].type);
var allAreObjectsOrLists = R__default['default'].compose(R__default['default'].either(R__default['default'].all(isObject), R__default['default'].all(Array.isArray)), R__default['default'].values);

var hasVersion = function hasVersion(versionLabel) {
  return function (obj) {
    return R__default['default'].has(versionLabel, obj);
  };
};

var hasBoth = function hasBoth(versionLabels) {
  return function (obj) {
    return R__default['default'].all(function (versionLabel) {
      return hasVersion(versionLabel)(obj);
    }, versionLabels);
  };
};

var isEqual = function isEqual(versionLabels) {
  return function (obj) {
    return R__default['default'].both(hasBoth(versionLabels), R__default['default'].compose(R__default['default'].apply(R__default['default'].equals), R__default['default'].values))(obj);
  };
};

var markAdded = function markAdded(obj) {
  return R__default['default'].compose(R__default['default'].append(undefined), R__default['default'].values)(obj);
}; // eslint-disable-line no-undefined


var markRemoved = function markRemoved(obj) {
  return R__default['default'].compose(R__default['default'].prepend(undefined), R__default['default'].values)(obj);
}; // eslint-disable-line no-undefined


var isAddition = function isAddition(versionLabels) {
  return function (obj) {
    return R__default['default'].both(hasVersion(R__default['default'].head(versionLabels)), R__default['default'].complement(hasVersion(R__default['default'].last(versionLabels))))(obj);
  };
};

var isRemoval = function isRemoval(versionLabels) {
  return function (obj) {
    return R__default['default'].both(R__default['default'].complement(hasVersion(R__default['default'].head(versionLabels))), hasVersion(R__default['default'].last(versionLabels)))(obj);
  };
};
/**
 * Diffs objects and replaces differences with an object {__left: left value, __right: right value}
 * @param {[String]} Two values indicating the names of the versions. If left null then __left and __right
 * are used for the diff
 * @param {Object} l The left object to diff
 * @param {Object} r The right object to diff
 * @returns {Object} A deep diff of l and r. Equal values are removed, different values are replaced by an
 * object keyed by the two version strings with the corresponding values
 */


var objectDiff = R__default['default'].curry(function (versionLabels, l, r) {
  var labels = R__default['default'].defaultTo(['__left', '__right'], versionLabels);

  var _labels = functions._slicedToArray(labels, 2),
      left = _labels[0],
      right = _labels[1];

  return R__default['default'].compose(function (result) {
    return R__default['default'].map(R__default['default'].cond([[function (obj) {
      return isAddition(labels)(obj);
    }, function (obj) {
      return markAdded(obj);
    }], [function (obj) {
      return isRemoval(labels)(obj);
    }, function (obj) {
      return markRemoved(obj);
    }], [function (obj) {
      return hasBoth(labels)(obj);
    }, function (obj) {
      return R__default['default'].when(allAreObjectsOrLists, // If all are objects or list recurse
      // Lists necessary become objects keyed by index since some indices are removed
      function (o) {
        return R__default['default'].compose( // Remove the left, right labels and recurse
        function (values) {
          return R__default['default'].apply(objectDiff(versionLabels), values);
        }, R__default['default'].values)(o);
      } // Otherwise we keep the left and right labels and we are done
      )(obj);
    }]]), result);
  }, function (result) {
    return R__default['default'].reject(function (obj) {
      return isEqual(labels)(obj);
    }, result);
  }, R__default['default'].useWith(R__default['default'].mergeWith(R__default['default'].merge), [R__default['default'].map(R__default['default'].objOf(left)), R__default['default'].map(R__default['default'].objOf(right))]))(l, r);
});
/**
 * Pretty-prints objectDiff
 * @param {[String]} versionLabels The two labels or null
 * @param {Object} l The left-side value
 * @param {Object} r The right-side value
 * @returns {String} The diff string
 */

var prettyPrintObjectDiff = function prettyPrintObjectDiff(versionLabels, l, r) {
  return R__default['default'].compose(function (result) {
    return JSON.stringify(result, null, 2);
  }, objectDiff)(versionLabels, l, r);
};

/**
 * Created by Andy Likuski on 2017.06.06
 * Copyright (c) 2017 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the 'Software'), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/**
 * Given a task, wraps it in promise and passes it to Jest's expect.
 * With this you can call resolves or rejects depending on whether success or failure is expected:
 * expectTask(task).resolves|rejects
 * @param {Task} task Task wrapped in a Promise and forked
 * @returns {undefined}
 */

var expectTask = function expectTask(task) {
  return expect(monadHelpers.taskToPromise(task));
};
/**
 * Converts an Result to a Promise. Result.Ok calls resolve and Result.Error calls reject
 * @param {Object} result A Result
 * @returns {Promise} the promise
 */

var resultToPromise = function resultToPromise(result) {
  return new Promise(function (resolve, reject) {
    return result.map(resolve).mapError(reject);
  });
};
/**
 * Convenient way to check if an object has a few expected keys at the given path
 * @param {[String]} keyPaths keys or dot-separated key paths of the object to check
 * @param {Object} obj The object to check
 * @return {*} Expects the object has the given keys. Throws if expect fails* @return {*}
 */

var expectKeys = R__default['default'].curry(function (keyPaths, obj) {
  expect(R__default['default'].compose( // Put the keyPaths that survive in a set for comparison
  function (a) {
    return new Set(a);
  }, // Filter out keyPaths that don't resolve to a non-nil value
  function (o) {
    return R__default['default'].filter(function (keyPath) {
      return R__default['default'].complement(R__default['default'].isNil)(R__default['default'].view(R__default['default'].lensPath(functions.keyStringToLensPath(keyPath)), o));
    }, keyPaths);
  })(obj)).toEqual(new Set(keyPaths)); // Required for validated functions

  return true;
});
/**
 * Convenient way to check if an object has a few expected keys at the given path
 * @param {[String]} keyPaths keys or dot-separated key paths of the object to check
 * @param {String} strPath path in the obj to check keyPaths for
 * @param {Object} obj The object to check
 * @return {*} Expects the object has the given keys. Throws if expect fails* @return {*}
 */

var expectKeysAtPath = R__default['default'].curry(function (keyPaths, strPath, obj) {
  expect(R__default['default'].compose( // Put the keyPaths that survive in a set for comparison
  function (a) {
    return new Set(a);
  }, // Filter out keyPaths that don't resolve to a non-nil value
  function (o) {
    return R__default['default'].filter(function (keyPath) {
      return R__default['default'].complement(R__default['default'].isNil)(R__default['default'].view(R__default['default'].lensPath(functions.keyStringToLensPath(keyPath)), throwingFunctions.reqStrPathThrowing(strPath, o)));
    }, keyPaths);
  })(obj)).toEqual(new Set(keyPaths)); // Required for validated functions

  return true;
});

exports.alwaysFunc = functions.alwaysFunc;
exports.applyDeep = functions.applyDeep;
exports.applyDeepWithKeyWithRecurseArraysAndMapObjs = functions.applyDeepWithKeyWithRecurseArraysAndMapObjs;
exports.camelCase = functions.camelCase;
exports.capitalize = functions.capitalize;
exports.chainObjToValues = functions.chainObjToValues;
exports.compact = functions.compact;
exports.compactEmpty = functions.compactEmpty;
exports.compactJoin = functions.compactJoin;
exports.duplicateKey = functions.duplicateKey;
exports.emptyToNull = functions.emptyToNull;
exports.eqStrPath = functions.eqStrPath;
exports.eqStrPathsAll = functions.eqStrPathsAll;
exports.eqStrPathsAllCustomizable = functions.eqStrPathsAllCustomizable;
exports.filterObjToValues = functions.filterObjToValues;
exports.filterWithKeys = functions.filterWithKeys;
exports.findByParams = functions.findByParams;
exports.findMapped = functions.findMapped;
exports.findOne = functions.findOne;
exports.findOneValueByParams = functions.findOneValueByParams;
exports.flattenObj = functions.flattenObj;
exports.flattenObjUntil = functions.flattenObjUntil;
exports.fromPairsDeep = functions.fromPairsDeep;
exports.hasStrPath = functions.hasStrPath;
exports.idOrIdFromObj = functions.idOrIdFromObj;
exports.isObject = functions.isObject;
exports.keyStringToLensPath = functions.keyStringToLensPath;
exports.lowercase = functions.lowercase;
exports.mapDefault = functions.mapDefault;
exports.mapDefaultAndPrefixOthers = functions.mapDefaultAndPrefixOthers;
exports.mapKeys = functions.mapKeys;
exports.mapKeysAndValues = functions.mapKeysAndValues;
exports.mapKeysForLens = functions.mapKeysForLens;
exports.mapObjToValues = functions.mapObjToValues;
exports.mapProp = functions.mapProp;
exports.mapPropValueAsIndex = functions.mapPropValueAsIndex;
exports.mapToObjValue = functions.mapToObjValue;
exports.mergeAllWithKey = functions.mergeAllWithKey;
exports.mergeDeep = functions.mergeDeep;
exports.mergeDeepAll = functions.mergeDeepAll;
exports.mergeDeepWith = functions.mergeDeepWith;
exports.mergeDeepWithConcatArrays = functions.mergeDeepWithConcatArrays;
exports.mergeDeepWithKeyWithRecurseArrayItemsAndMapObjs = functions.mergeDeepWithKeyWithRecurseArrayItemsAndMapObjs;
exports.mergeDeepWithRecurseArrayItems = functions.mergeDeepWithRecurseArrayItems;
exports.mergeDeepWithRecurseArrayItemsByAndMergeObjectByRight = functions.mergeDeepWithRecurseArrayItemsByAndMergeObjectByRight;
exports.mergeDeepWithRecurseArrayItemsByRight = functions.mergeDeepWithRecurseArrayItemsByRight;
exports.moveToKeys = functions.moveToKeys;
exports.omitDeep = functions.omitDeep;
exports.omitDeepBy = functions.omitDeepBy;
exports.omitDeepPaths = functions.omitDeepPaths;
exports.onlyOne = functions.onlyOne;
exports.onlyOneValue = functions.onlyOneValue;
exports.orEmpty = functions.orEmpty;
exports.overDeep = functions.overDeep;
exports.pickDeepPaths = functions.pickDeepPaths;
exports.removeDuplicateObjectsByProp = functions.removeDuplicateObjectsByProp;
exports.renameKey = functions.renameKey;
exports.replaceValuesAtDepth = functions.replaceValuesAtDepth;
exports.replaceValuesAtDepthAndStringify = functions.replaceValuesAtDepthAndStringify;
exports.replaceValuesWithCountAtDepth = functions.replaceValuesWithCountAtDepth;
exports.replaceValuesWithCountAtDepthAndStringify = functions.replaceValuesWithCountAtDepthAndStringify;
exports.reqPath = functions.reqPath;
exports.reqPathPropEq = functions.reqPathPropEq;
exports.reqStrPath = functions.reqStrPath;
exports.splitAtInclusive = functions.splitAtInclusive;
exports.strPath = functions.strPath;
exports.strPathOr = functions.strPathOr;
exports.strPathOrNullOk = functions.strPathOrNullOk;
exports.strPathsOr = functions.strPathsOr;
exports.strPathsOrNullOk = functions.strPathsOrNullOk;
exports.toArrayIfNot = functions.toArrayIfNot;
exports.transformKeys = functions.transformKeys;
exports.unflattenObj = functions.unflattenObj;
exports.unflattenObjNoArrays = functions.unflattenObjNoArrays;
exports.findOneThrowing = throwingFunctions.findOneThrowing;
exports.findOneValueByParamsThrowing = throwingFunctions.findOneValueByParamsThrowing;
exports.mappedThrowIfResultError = throwingFunctions.mappedThrowIfResultError;
exports.onlyOneThrowing = throwingFunctions.onlyOneThrowing;
exports.onlyOneValueThrowing = throwingFunctions.onlyOneValueThrowing;
exports.reqPathPropEqThrowing = throwingFunctions.reqPathPropEqThrowing;
exports.reqPathThrowing = throwingFunctions.reqPathThrowing;
exports.reqStrPathThrowing = throwingFunctions.reqStrPathThrowing;
exports.throwIfResultError = throwingFunctions.throwIfResultError;
exports.throwIfSingleResultError = throwingFunctions.throwIfSingleResultError;
exports.chainExceptMapDeepestMDeep = monadHelpers.chainExceptMapDeepestMDeep;
exports.chainMDeep = monadHelpers.chainMDeep;
exports.composeWithChain = monadHelpers.composeWithChain;
exports.composeWithChainMDeep = monadHelpers.composeWithChainMDeep;
exports.composeWithMap = monadHelpers.composeWithMap;
exports.composeWithMapExceptChainDeepestMDeep = monadHelpers.composeWithMapExceptChainDeepestMDeep;
exports.composeWithMapMDeep = monadHelpers.composeWithMapMDeep;
exports.defaultRunConfig = monadHelpers.defaultRunConfig;
exports.defaultRunToResultConfig = monadHelpers.defaultRunToResultConfig;
exports.doMDeep = monadHelpers.doMDeep;
exports.doMDeepExceptDeepest = monadHelpers.doMDeepExceptDeepest;
exports.lift1stOf2ForMDeepMonad = monadHelpers.lift1stOf2ForMDeepMonad;
exports.mapExceptChainDeepestMDeep = monadHelpers.mapExceptChainDeepestMDeep;
exports.mapMDeep = monadHelpers.mapMDeep;
exports.mapMonadByConfig = monadHelpers.mapMonadByConfig;
exports.mapOrObjToNamedResponseAndInputs = monadHelpers.mapOrObjToNamedResponseAndInputs;
exports.mapResultMonadWithOtherInputs = monadHelpers.mapResultMonadWithOtherInputs;
exports.mapResultTaskWithOtherInputs = monadHelpers.mapResultTaskWithOtherInputs;
exports.mapToMergedResponseAndInputs = monadHelpers.mapToMergedResponseAndInputs;
exports.mapToMergedResponseAndInputsMDeep = monadHelpers.mapToMergedResponseAndInputsMDeep;
exports.mapToNamedPathAndInputs = monadHelpers.mapToNamedPathAndInputs;
exports.mapToNamedResponseAndInputs = monadHelpers.mapToNamedResponseAndInputs;
exports.mapToNamedResponseAndInputsMDeep = monadHelpers.mapToNamedResponseAndInputsMDeep;
exports.mapToPath = monadHelpers.mapToPath;
exports.mapToResponseAndInputs = monadHelpers.mapToResponseAndInputs;
exports.mapWithArgToPath = monadHelpers.mapWithArgToPath;
exports.objOfMLevelDeepListOfMonadsToListWithPairs = monadHelpers.objOfMLevelDeepListOfMonadsToListWithPairs;
exports.objOfMLevelDeepMonadsToListWithPairs = monadHelpers.objOfMLevelDeepMonadsToListWithPairs;
exports.pairsOfMLevelDeepListOfMonadsToListWithPairs = monadHelpers.pairsOfMLevelDeepListOfMonadsToListWithPairs;
exports.promiseToTask = monadHelpers.promiseToTask;
exports.resultTasksToResultObjTask = monadHelpers.resultTasksToResultObjTask;
exports.resultToTask = monadHelpers.resultToTask;
exports.resultToTaskNeedingResult = monadHelpers.resultToTaskNeedingResult;
exports.resultToTaskWithResult = monadHelpers.resultToTaskWithResult;
exports.resultsToResultObj = monadHelpers.resultsToResultObj;
exports.retryTask = monadHelpers.retryTask;
exports.sequenceBucketed = monadHelpers.sequenceBucketed;
exports.stringifyError = monadHelpers.stringifyError;
exports.taskToPromise = monadHelpers.taskToPromise;
exports.taskToResultTask = monadHelpers.taskToResultTask;
exports.toMergedResponseAndInputs = monadHelpers.toMergedResponseAndInputs;
exports.toNamedResponseAndInputs = monadHelpers.toNamedResponseAndInputs;
exports.traverseReduce = monadHelpers.traverseReduce;
exports.traverseReduceDeep = monadHelpers.traverseReduceDeep;
exports.traverseReduceDeepResults = monadHelpers.traverseReduceDeepResults;
exports.traverseReduceError = monadHelpers.traverseReduceError;
exports.traverseReduceResultError = monadHelpers.traverseReduceResultError;
exports.traverseReduceWhile = monadHelpers.traverseReduceWhile;
exports.traverseReduceWhileBucketed = monadHelpers.traverseReduceWhileBucketed;
exports.traverseReduceWhileBucketedTasks = monadHelpers.traverseReduceWhileBucketedTasks;
exports.waitAllBucketed = monadHelpers.waitAllBucketed;
exports.expectKeys = expectKeys;
exports.expectKeysAtPath = expectKeysAtPath;
exports.expectTask = expectTask;
exports.memoized = memoized;
exports.memoizedWith = memoizedWith;
exports.objectDiff = objectDiff;
exports.prettyPrintObjectDiff = prettyPrintObjectDiff;
exports.resultToPromise = resultToPromise;
//# sourceMappingURL=index.js.map
